//	Ported from:
//
//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "internal.h"

class BlendFields : public GenericVideoFilter {
public:
	BlendFields(PClip _child, IScriptEnvironment *env)
		: GenericVideoFilter(_child)
	{
		if (!vi.IsYUY2() && !vi.IsRGB())
            env->ThrowError("BlendFields: requires RGB or YUY2 data");
		if (vi.IsFieldBased()) {
			PClip c = env->Invoke("Weave",_child).AsClip();
			child = c;
			vi = c->GetVideoInfo();
		}
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

static void __declspec(naked) asm_blend_row_clipped(unsigned char *dst, const unsigned char *src, int w, int srcpitch) {
	__asm {
		push	ebp
		push	edi
		push	esi
		push	ebx

		mov		edi,[esp+20]
		mov		esi,[esp+24]
		sub		edi,esi
		mov		ebp,[esp+28]
		mov		edx,[esp+32]

xloop:
		mov		ecx,[esi]
		mov		eax,0fefefefeh

		mov		ebx,[esi+edx]
		and		eax,ecx

		shr		eax,1
		and		ebx,0fefefefeh

		shr		ebx,1
		add		esi,4

		add		eax,ebx
		dec		ebp

		mov		[edi+esi-4],eax
		jnz		xloop

		pop		ebx
		pop		esi
		pop		edi
		pop		ebp
		ret
	};
}

static void __declspec(naked) asm_blend_row(unsigned char *dst, const unsigned char *src, int w, int srcpitch) {
	__asm {
		push	ebp
		push	edi
		push	esi
		push	ebx

		mov		edi,[esp+20]
		mov		esi,[esp+24]
		sub		edi,esi
		mov		ebp,[esp+28]
		mov		edx,[esp+32]

xloop:
		mov		ecx,[esi]
		mov		eax,0fcfcfcfch

		mov		ebx,[esi+edx]
		and		eax,ecx

		shr		ebx,1
		mov		ecx,[esi+edx*2]

		shr		ecx,2
		and		ebx,07f7f7f7fh

		shr		eax,2
		and		ecx,03f3f3f3fh

		add		eax,ebx
		add		esi,4

		add		eax,ecx
		dec		ebp

		mov		[edi+esi-4],eax
		jnz		xloop

		pop		ebx
		pop		esi
		pop		edi
		pop		ebp
		ret
	};
}

#pragma warning(disable:4799)

static void __declspec(naked) asm_blend_row_MMX(unsigned char *dst, const unsigned char *src, int w, int srcpitch) {
	static const __int64 mask0 = 0xfcfcfcfcfcfcfcfci64;
	static const __int64 mask1 = 0x7f7f7f7f7f7f7f7fi64;
	static const __int64 mask2 = 0x3f3f3f3f3f3f3f3fi64;
	__asm {
		push	ebp
		push	edi
		push	esi
		push	ebx

		mov		edi,[esp+20]
		mov		esi,[esp+24]
		sub		edi,esi
		mov		ebp,[esp+28]
		mov		edx,[esp+32]

		movq	mm5,mask0
		movq	mm6,mask1
		movq	mm7,mask2
		shr		ebp,1
		jz		oddpart

xloop:
		movq	mm2,[esi]
		movq	mm0,mm5

		movq	mm1,[esi+edx]
		pand	mm0,mm2

		psrlq	mm1,1
		movq	mm2,[esi+edx*2]

		psrlq	mm2,2
		pand	mm1,mm6

		psrlq	mm0,2
		pand	mm2,mm7

		paddb	mm0,mm1
		add		esi,8

		paddb	mm0,mm2
		dec		ebp

		movq	[edi+esi-8],mm0
		jne		xloop

oddpart:
		test	byte ptr [esp+28],1
		jz		nooddpart

		mov		ecx,[esi]
		mov		eax,0fcfcfcfch
		mov		ebx,[esi+edx]
		and		eax,ecx
		shr		ebx,1
		mov		ecx,[esi+edx*2]
		shr		ecx,2
		and		ebx,07f7f7f7fh
		shr		eax,2
		and		ecx,03f3f3f3fh
		add		eax,ebx
		add		eax,ecx
		mov		[edi+esi],eax

nooddpart:

		pop		ebx
		pop		esi
		pop		edi
		pop		ebp
		ret
	};
}

#pragma warning(default:4799)


PVideoFrame __stdcall BlendFields::GetFrame(int n, IScriptEnvironment* env)
{

    PVideoFrame src = child->GetFrame(n, env);
    int h= src->GetHeight();

	if (h<2) return src;

    PVideoFrame dst = env->NewVideoFrame(vi);

    const unsigned char* srcp = src->GetReadPtr();
    unsigned char* dstp = dst->GetWritePtr();

    const int src_pitch = src->GetPitch();
    const int dst_pitch = dst->GetPitch();
    const int w= dst->GetRowSize()>>2;
	void (*blend_func)(unsigned char *,const unsigned char *, int, int);

	blend_func = (env->GetCPUFlags() & CPUF_MMX) ? asm_blend_row_MMX : asm_blend_row;

	asm_blend_row_clipped(dstp, srcp, w, src_pitch);
	if (h-=2)
		do {
			dstp = dstp + dst_pitch;
			blend_func(dstp, srcp, w, src_pitch);
			srcp = srcp + src_pitch;
		} while(--h);

	asm_blend_row_clipped(dstp + dst_pitch, srcp, w, src_pitch);

	if (env->GetCPUFlags() & CPUF_MMX)
		__asm emms

	return dst;
}


AVSValue __cdecl Create_BlendFields(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new BlendFields(args[0].AsClip(), env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("BlendFields", "c", Create_BlendFields, 0);
    return 0;
}
